import json
import os
import random
import requests
import string
import time
from orka_sdk.orka_sdk import OrkaSDK

class GHAController:

	def __init__(self):
		self.orka_user = os.environ.get('ORKA_USER')
		self.password = os.environ.get('ORKA_PASS')
		self.license_key = os.environ.get('ORKA_LICENSE_KEY')
		self.github_user = os.environ.get('GH_USER')
		self.github_pat = os.environ.get('GH_PAT')
		self.github_repo_name = os.environ.get('GH_REPO')
		self.core_count = os.environ.get('CORE_COUNT')
		self.vcpu_count = os.environ.get('VCPU_COUNT')
		self.base_image = os.environ.get('ORKA_BASE_IMAGE')
		self.gh_session = requests.Session()
		self.gh_session.auth = (self.github_user, self.github_pat)
		self.orka = OrkaSDK()
		self.orka.login(
			user=self.orka_user, 
			password=self.password, 
			license_key=self.license_key)
		self.vm = None
		self.vm_name = None
		self.runner_id = None

	def _generate_runner_name(self):
		pool = string.ascii_lowercase
		self.vm_name = ''.join(random.choice(pool) for i in range(10))

	def spin_up(self):
		self._generate_runner_name()
		if not self.core_count:
			self.core_count = '6'
		if not self.vcpu_count:
			self.vcpu_count = '6'
		vm_data = {
			'vm_name': self.vm_name,
			'orka_base_image': self.base_image,
			'core_count': self.core_count,
			'vcpu_count': self.vcpu_count
		}

		vm_metadata = {
			'github_user': self.github_user,
			'github_pat': self.github_pat,
			'github_repo_name': self.github_repo_name,
		}

		r = self.orka.create_vm_config(vm_data)
		if r.errors:
			for e in r.errors:
				print(e)
		r = self.orka.deploy_vm_config(
			vm_name=vm_data['vm_name'], 
			vm_metadata=vm_metadata)
		if r.errors:
			for e in r.errors:
				print(e)
		self.vm = r.data

	def check_runner_status(self):
		headers = {'Accept':'application/vnd.github.v3+json'}
		url = (f'https://api.github.com/repos/'
			f'{self.github_user}/{self.github_repo_name}/actions/runners')
		result = self.gh_session.get(url, headers=headers)
		content = json.loads(result._content.decode('utf-8'))
		for item in content['runners']:
			if self.vm_name in item['name']:
				return True
		else:
			time.sleep(4)
			# relay logs
			r = self.vm.exec('cat /Users/admin/runner-connect.log')
			print(r.data['stdout'])
			self.check_runner_status()

	def tear_down(self):
		if not self.vm_name:
			self.vm_name = self.vm.name
		self.orka.purge_vm(self.vm)
		self._get_runner_id()
		self._remove_runner_from_gh()

	def _get_runner_id(self):
		headers = {'Accept':'application/vnd.github.v3+json'}
		url = (f'https://api.github.com/repos/{self.github_user}'
			f'/{self.github_repo_name}/actions/runners')
		result = self.gh_session.get(url, headers=headers)
		content = json.loads(result._content.decode('utf-8'))
		for item in content['runners']:
			if self.vm_name in item['name']:
				self.runner_id = item['id']

	def _remove_runner_from_gh(self):
		headers = {'Accept':'application/vnd.github.v3+json'}
		url = (f'https://api.github.com/repos/{self.github_user}'
			f'/{self.github_repo_name}/actions/runners/{self.runner_id}')
		self.gh_session.delete(url,headers=headers)



import os
import paramiko
from jinja2 import Environment
from jinja2 import FileSystemLoader
from orka_sdk.result import Result

class VM():

	def __init__(self, data):
		self.ip = data['ip']
		self.name = data['name']
		self.ssh_port = data['ssh_port']
		self.id = data['id']
		self.ram = data['ram']
		self.vcpu = data['vcpu']
		self.cpu = data['cpu']
		self.io_boost = data['io_boost']
		self.use_saved_state = data['use_saved_state']
		self.gpu_passthrough = data['gpu_passthrough']
		self.screen_share_port = data['screen_share_port']
		self.vnc_port = data['vnc_port']
		self.ssh_client = paramiko.SSHClient()
		self.ssh_client.set_missing_host_key_policy(
			paramiko.AutoAddPolicy())
		self.sftp_client = None
		self.ssh_user = 'admin'
		self.ssh_pass = 'admin'

	def _connect_ssh_client(self):
		self.ssh_client.connect(
		self.ip, username=self.ssh_user, 
		password=self.ssh_pass, 
		port=self.ssh_port, 
		look_for_keys=False, 
		timeout=60, 
		allow_agent=False)

	def exec(self, cmd):
		output = {}
		try:
			self._connect_ssh_client()
		except Exception as e:

			return Result(errors=[str(e)])
		try:	
			stdin, stdout, stderr = \
				self.ssh_client.exec_command(cmd)
			stdin.write(f'{self.ssh_pass}\n')
			stdin.flush()
		except Exception as e:

			return Result(errors=[str(e)])

		output['stdout'] = stdout.read().decode("utf8")
		output['stderr'] = stderr.read().decode("utf8")
		stdin.close()
		stdout.close()
		stderr.close()

		return Result(errors=None, data=output)

	def _create_sftp_client(self):
		# only to be called after _connect_ssh_client()
		self.sftp_client = self.ssh_client.open_sftp()

	def upload(self, local_path, dest_path):
		try:
			self._connect_ssh_client()
			self._create_sftp_client()
		except Exception as e:

			return Result(errors=[str(e)])
		try:
			self.sftp_client.put(
				local_path, dest_path)
		except Exception as e:

			return Result(errors=[str(e)])

		return Result(errors=None)

	def download(self, dest_path, local_path):
		try:
			self._connect_ssh_client()
			self._create_sftp_client()
		except Exception as e:

			return Result(errors=[str(e)])
		try:
			self.sftp_client.get(
				dest_path, local_path)
		except Exception as e:

			return Result(errors=[str(e)])

		return Result(errors=None)

	def write_persistent_env_var(self, data, dest=None):
		dest_text = ''
		exports = []
		if not dest:
			dest = '/Users/admin/.zshenv'

		for var, value in data.items():
			export = f'export {var}={value}\n'
			exports.append(export)

		# if the file exists already, read its contents
		r = self.exec(f'cat {dest}')
		if r.errors:

			return r
		if r.data['stdout']:
			dest_text = r.data['stdout']

		# write the temp file locally
		with open('environment.temp', 'w') as f:
			f.write(dest_text)
			# sloppily append exports to end for now
			for export in exports:
				f.write(export)
		r = self.upload('environment.temp', dest)

		# clean up
		self._remove_temp_files()

		return r

	def enable_auto_login(self):
		cmd = (f'sudo -S -p "" defaults write' 
			f' /Library/Preferences/com.apple.loginwindow.plist' 
			f' autoLoginUser {self.ssh_user}')
		r = self.exec(cmd)
		if r.errors:

			return r
		cmd = (f'sudo -S -p "" plutil -replace autoLoginUser -string'
			f' {self.ssh_user} /Library/Preferences/com.apple.loginwindow.plist')
		r = self.exec(cmd)
		if r.errors:

			return r
		cmd = (f'sudo -S -p "" /usr/libexec/PlistBuddy -c "Set autoLoginUser' 
			f' {self.ssh_user}" /Library/Preferences/com.apple.loginwindow.plist')
		r = self.exec(cmd)

		return r

	def create_launch_daemon(self, data):
		# render template as temp file
		cwd = os.getcwd()
		env = Environment(loader=FileSystemLoader(cwd))
		template = env.get_template('launch_daemon.jinja')
		with open('launch_daemon.temp', 'w') as f:
			f.write(template.render(data))

		# upload temp file
		r = self.upload(
			'launch_daemon.temp',
			'/tmp/launch_daemon.temp')
		if r.errors:

			return r

		# move file with sudo
		r = self.exec(
			f'sudo -S -p "" mv /tmp/launch_daemon.temp ' 
			f'/Library/LaunchDaemons/com.{data["name"]}.app.plist')

		# clean up
		self._remove_temp_files()
		try:
			r = self.exec('rm -rf /tmp/launch_daemon.temp')
		except Exception as e:
			print('Warning: failed to remove remote temp file.')

			return r

		return r

	def _remove_temp_files(self):
		cwd = os.getcwd()
		for f in os.listdir(cwd):
			if '.temp' in f:
				path = os.path.join(cwd, f)
				os.remove(path)

		return None



import os
import requests
import yaml
from pathlib import Path
from kubernetes import client, config
from orka_sdk.result import Result


class K8s:

	def __init__(self, base_sdk, config_path=None):
		self.token = base_sdk.token
		self.user = base_sdk.user
		self.password = base_sdk.password
		self.license_key = base_sdk.license_key
		self.orka_ip = base_sdk.orka_ip
		if config_path:
			try:
				config.load_kube_config(config_path)
				self.client = client.AppsV1Api()

				return None
			except Exception as e:
				errors = [str(e)]
				print('Failed to create Kubernetes client.')
				print(errors[0])
				self.client = None
				
				return None
		home = str(Path.home())
		config_path = \
			os.path.join(home, 'kubeconfig-orka')
		if os.path.exists(config_path):
			config.load_kube_config(config_path)
			self.client = client.AppsV1Api()
		else:
			self.client = None

	def create_deployment(self, yaml_path):
		with open(yaml_path) as f:
			deployment = yaml.safe_load(f)
			try:
				r = self.client.create_namespaced_deployment(
					body=deployment, namespace='sandbox')
			except Exception as e:
				errors = [str(e)]

				return Result(errors=errors)
			
			return r

	def delete_deployment(self, name):
		try:
			r = self.client.delete_namespaced_deployment(
				name=name,
				namespace="sandbox",
				body=client.V1DeleteOptions(
					propagation_policy="Foreground",
					grace_period_seconds=5
				)
		  	)
		except Exception as e:
			errors = [str(e)]

			return Result(errors=errors)

		return r

	def create_service(self, yaml_path):
		with open(yaml_path) as f:
			service = yaml.safe_load(f)
			try:
				r = self.client.create_namespaced_service(
					body=service, namespace='sandbox')
			except Exception as e:
				errors = [str(e)]

				return Result(errors=errors)
			
			return r

	def delete_service(self, name):
		try:
			r = self.client.delete_namespaced_service(
				name=name,
				namespace="sandbox",
				body=client.V1DeleteOptions(
					propagation_policy="Foreground",
					grace_period_seconds=5
				)
		  	)
		except Exception as e:
			errors = [str(e)]

			return Result(errors=errors)

		return r


import json
import time
import requests
from .result import Result
from orka_sdk.vm import VM
from orka_sdk.images import Images
from orka_sdk.k8s import K8s
from orka_sdk.nodes import Nodes


ORKA_IP = 'http://10.221.188.100'

class OrkaSDK:

	def __init__(self):
		self.orka_ip = ORKA_IP
		self.token = None
		self.user = None
		self.password = None
		self.license_key = None
		self.k8s = None
		self.images = None
		self.nodes = None

###################  Auth  ###################

	def login(self, user, password, license_key=None):
		errors = []
		self.user = user
		self.password = password
		self.license_key = license_key
		try:
			self.token = self._get_token(user, password)
			self.nodes = Nodes(self)
			self.images = Images(self)
			if self.license_key:
				self.k8s = K8s(self)
			else:
				# TODO: logging?
				print('Kubernetes functionality disabled. Requires Orka license-key.')
		except Exception as e:
			errors = [str(e)]

			return Result(errors=errors)

		return Result(errors=errors)

	def _get_token(self, user, password):
		headers = {'Content-Type': 'application/json'}
		data = json.dumps({'email': user, 'password': password})
		result = requests.post(f"{ORKA_IP}/token", data=data, headers=headers)
		content = json.loads(result._content.decode('utf-8'))

		return content['token']

	def revoke_token(self):
		pass

###############  VM Management  ###############

	def create_vm(self, vm_data):
		self.create_vm_config(vm_data)

		return self.deploy_vm_config(vm_data['vm_name'])

	def create_vm_config(self, vm_data):
		url = f"{ORKA_IP}/resources/vm/create"
		headers = {
			'Content-Type': 'application/json', 
			'Authorization': f"Bearer {self.token}"
			}
		data = {
			'orka_vm_name': vm_data['vm_name'],
			'orka_base_image': vm_data['orka_base_image'],
			'orka_image': vm_data['vm_name'],
			'orka_cpu_core': int(vm_data['core_count']),
			'vcpu_count': int(vm_data['vcpu_count'])
			}
		r = requests.post(url, data=json.dumps(data), headers=headers)
		content = json.loads(r._content.decode('utf-8'))
		errors = content.get('errors')
		
		return Result(errors=errors)
        
	def deploy_vm_config(self, vm_name, vm_metadata=None):
		url = f"{ORKA_IP}/resources/vm/deploy"
		headers = {
			'Content-Type': 'application/json', 
			'Authorization': f"Bearer {self.token}"
			}
		data =  {'orka_vm_name': vm_name}
		if vm_metadata:
			data['vm_metadata'] = {}
			items = []
			for key, value in vm_metadata.items():
				item = '{'+f'"key":"{key}", "value":"{value}"'+'}'
				items.append(json.loads(item))

			data['vm_metadata']['items'] = items
		r = requests.post(url, data=json.dumps(data), headers=headers)
		content = json.loads(r._content.decode('utf-8'))
		errors = content.get('errors')
		if errors:

			return Result(errors=errors)
		time.sleep(15)
		try:
			data = self._parse_config_response(content)
			vm = VM(data)
		except Exception as e:
			errors = [str(e)]

			return Result(errors=errors)
		
		return Result(errors=errors, data=vm)

	def _parse_config_response(self, content):
		data = {}
		data['ip'] = content['ip']
		data['id'] = content['vm_id']
		data['ssh_port'] = content['ssh_port']
		data['name'] = \
			content['help']['data_for_virtual_machine_exec_tasks']['orka_vm_name']
		data['ram'] = content['ram']
		data['vcpu'] = content['vcpu']
		data['cpu'] = content['host_cpu']
		data['io_boost'] = content['io_boost']
		data['use_saved_state'] = content['use_saved_state']
		data['gpu_passthrough'] = content['gpu_passthrough']
		data['screen_share_port'] = content['screen_share_port']
		data['vnc_port'] = content['vnc_port']

		return data

	def list_session_vms(self):
		url = f"{ORKA_IP}/resources/vm/list"
		headers = {
			'Content-Type': 'application/json', 
			'Authorization': f"Bearer {self.token}"
			}
		r = requests.get(url, headers=headers)
		content = json.loads(r._content.decode('utf-8'))
		errors = content.get('errors')
		if errors:
		
			return Result(errors=errors)
		vm_instances = self._instantiate_vms(content)

		return Result(errors=errors, data=vm_instances)

	def list_user_vms(self, user=None):
		if self.license_key:
			if not user:
				user = self.user
			url = f"{ORKA_IP}/resources/vm/list/{user}"
			headers = {
				'Content-Type': 'application/json', 
				'Authorization': f"Bearer {self.token}",
				'orka-licensekey': self.license_key
				}
			r = requests.get(url, headers=headers)
			content = json.loads(r._content.decode('utf-8'))
			errors = content.get('errors')
			if errors:

				return Result(errors=errors)
			try:
				vm_instances = self._instantiate_vms(content)
			except Exception as e:
				errors = [f'Failed to instantiate VMs, missing key {str(e)}']

				return Result(errors=errors)

			return Result(errors=errors, data=vm_instances)
		else:
			errors = ['This method requires an Orka license-key']

			return Result(errors=errors)

	def list_system_vms(self):
		if self.license_key:
			url = f"{ORKA_IP}/resources/vm/list/all"
			headers = {
				'Content-Type': 'application/json', 
				'Authorization': f"Bearer {self.token}",
				'orka-licensekey': self.license_key
				}
			r = requests.get(url, headers=headers)
			content = json.loads(r._content.decode('utf-8'))
			errors = content.get('errors')
			if errors:

				return Result(errors=errors)
			try:
				vm_instances = self._instantiate_vms(content)
			except Exception as e:
				errors = [f'Failed to instantiate VMs, missing key {str(e)}']

				return Result(errors=errors)

			return Result(errors=None, data=vm_instances)
		else:
			errors = ['This method requires an orka license_key']

			return Result(errors=errors)

	def _instantiate_vms(self, content):
		vm_instances = []
		for vm in content['virtual_machine_resources']:
			if vm['vm_deployment_status'] == 'Deployed':
				data = {}
				data['ssh_port'] = vm['status'][0]['ssh_port']
				data['ip'] = vm['status'][0]['virtual_machine_ip']
				data['id'] = vm['status'][0]['virtual_machine_id']
				data['name'] = vm['virtual_machine_name']
				data['ram'] = vm['status'][0]['RAM']
				data['vcpu'] = vm['status'][0]['vcpu']
				data['cpu'] = vm['status'][0]['cpu']
				data['io_boost'] = vm['status'][0]['io_boost']
				data['use_saved_state'] = vm['status'][0]['use_saved_state']
				data['gpu_passthrough'] = vm['status'][0]['gpu']
				data['screen_share_port'] = vm['status'][0]['screen_sharing_port']
				data['vnc_port'] = vm['status'][0]['vnc_port']
				data['deployed'] = True
				vm = VM(data)
				vm_instances.append(vm)
			else:
				data = {}
				data['ssh_port'] = None
				data['ip'] = None
				data['id'] = None
				data['name'] = vm['virtual_machine_name']
				data['ram'] = None
				data['vcpu'] = vm['vcpu']
				data['cpu'] = vm['cpu']
				data['io_boost'] = vm['io_boost']
				data['use_saved_state'] = vm['use_saved_state']
				data['gpu_passthrough'] = vm['gpu_passthrough']
				data['screen_share_port'] = None
				data['vnc_port'] = None
				data['deployed'] = False
				vm = VM(data)
				vm_instances.append(vm)

		return vm_instances

	def get_vm_by_id(self, vm_id):
		r = self.list_system_vms()
		if r.errors:

			return Result(errors=r.errors)

		for vm in r.data:
			if vm_id == vm.id:
				data = vm

		if not data:
			errors = f'VM with id: {vm_id} not found'

			return Result(errors=errors)

		return Result(errors=None, data=data)

	def get_vm_by_name(self, name):
		pass

	def delete_vm(self, vm):
		url = f'{ORKA_IP}/resources/vm/delete'
		headers = {
		  'Content-Type': 'application/json',
		  'Authorization': f'Bearer {self.token}'
		}
		data = json.dumps({
		  'orka_vm_name': vm.name
		})
		r = requests.delete(url, headers=headers, data=data)
		content = json.loads(r._content.decode('utf-8'))
		errors = content.get('errors')
		
		return Result(errors=errors)

	def purge_vm(self, vm):
		url = f'{ORKA_IP}/resources/vm/purge'
		headers = {
		  'Content-Type': 'application/json',
		  'Authorization': f'Bearer {self.token}'
		}
		data = json.dumps({
		  'orka_vm_name': vm.name
		})
		r = requests.delete(url, headers=headers, data=data)
		content = json.loads(r._content.decode('utf-8'))
		errors = content.get('errors')
		
		return Result(errors=errors)


############# Image Management ###############

	def save_vm_as_image(self, image_name, vm):
		url = f'{ORKA_IP}/resources/image/save'
		headers = {
			'Content-Type': 'application/json', 
			'Authorization': f"Bearer {self.token}"
			}
		data = json.dumps({
			'orka_vm_name': vm.id,
			'new_name': image_name
		})
		r = requests.post(url, headers=headers, data=data)
		content = json.loads(r._content.decode('utf-8'))
		errors = content.get('errors')
		
		return Result(errors=errors)

	def commit_vm_state_to_base_image(self, vm):
		url = f'{ORKA_IP}/resources/image/commit'
		headers = {
		  'Content-Type': 'application/json',
		  'Authorization': f'Bearer {self.token}'
		}
		data = json.dumps({
		  'orka_vm_name': vm.id
		})
		r = requests.post(url, headers=headers, data=data)
		content = json.loads(r._content.decode('utf-8'))
		errors = content.get('errors')
		
		return Result(errors=errors)



from orka_sdk.result import Result


class Images:

	def __init__(self, base_sdk):
		self.token = base_sdk.token
		self.user = base_sdk.user
		self.password = base_sdk.password
		self.license_key = base_sdk.license_key
		self.orka_ip = base_sdk.orka_ip

	def list(self):
		url = f'{self.orka_ip}/resources/image/list'
		headers = {
			'Authorization': f'Bearer {self.token}'
			}
		r = requests.get(url, headers=headers)
		content = json.loads(r._content.decode('utf-8'))
		errors = content.get('errors')
		if errors:

			return Result(errors=errors)
		data = content['image_attributes']
		
		return Result(errors=errors, data=data)


import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="orka-sdk",
    version="1.0.2",
    description="Control Orka clusters and the macOS VMs they run.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/jeff-vincent/orka-python-sdk",
    author="Jeff Vincent",
    author_email="jeff.d.vincent@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    packages=["orka_sdk"],
    include_package_data=True,
    install_requires=[
        "attrs==21.4.0",
        "bcrypt==3.2.0",
        "cachetools==4.2.4",
        "certifi==2021.10.8",
        "cffi==1.15.0",
        "charset-normalizer==2.0.9",
        "cryptography==36.0.0",
        "google-auth==2.3.3",
        "idna==3.3",
        "importlib-metadata==4.10.1",
        "iniconfig==1.1.1",
        "Jinja2==3.0.3",
        "kubernetes==21.7.0",
        "MarkupSafe==2.0.1",
        "mock==4.0.3",
        "oauthlib==3.1.1",
        "packaging==21.3",
        "paramiko==2.8.1",
        "pluggy==1.0.0",
        "py==1.11.0",
        "pyasn1==0.4.8",
        "pyasn1-modules==0.2.8",
        "pycparser==2.21",
        "PyNaCl==1.4.0",
        "pyparsing==3.0.7",
        "pytest==6.2.5",
        "python-dateutil==2.8.2",
        "PyYAML==6.0",
        "requests==2.26.0",
        "requests-oauthlib==1.3.0",
        "rsa==4.8",
        "six==1.16.0",
        "toml==0.10.2",
        "typing_extensions==4.0.1",
        "urllib3==1.26.7",
        "websocket-client==1.2.3",
        "zipp==3.7.0"
        ]
)